from pymongo import MongoClient, errors
from .util import MONGO_BASE_HOST

class UserBuckets:
	def __init__(self, helm_release = None, mongo = None):
		if mongo is None:
			if helm_release is None:
				raise Exception('You must provide either monge or helm_release!')
			self._host ='%s-%s'%(helm_release, MONGO_BASE_HOST)
		else:
			self._host = mongo
		self._client = self._build_client(self._host)

	def _build_client(self, host):
		return MongoClient(host,
			connectTimeoutMS = 5000,
			serverSelectionTimeoutMS = 5000)

	def _pull_buckets(self):
		try:
			if not 'metadata' in self._client.database_names():
				# log_err('"metadata" collection not found in mongo!')
				sys.exit(1)
			db = self._client['metadata']
			if not '__usersbucket' in db.collection_names():
				# No buckets have been created or somethings wrong!
				return []
				sys.exit(1)
			col = db['__usersbucket']
			for bucket in col.find():
				yield bucket['_id'].split('..|..')[1]
		except errors.ServerSelectionTimeoutError:
			# log_err('Unable to connect to mongo!')
			sys.exit(1)

	@property
	def buckets(self):
		for name in self._pull_buckets():
			yield ZenkoBucket(self._client, name)


class ZenkoBucket:
	def __init__(self, client, name):
		self._name = name
		self._pulled = False
		self._client = client

	def _pull_info(self):
		if not self._pulled:
			try:
				if not 'metadata' in self._client.database_names():
					# log_err('"metadata" collection not found in mongo!')
					sys.exit(1)
				db = self._client['metadata']
				if not '__metastore' in db.collection_names():
					# log_err('No collection named PENSIEVE in metadata')
					sys.exit(1)
				col = db['__metastore']
				bucket = col.find_one({ '_id': self._name})
				if not bucket:
					# Could not find bucket collection
					sys.exit(1)
				self._pulled = True
				self._parse_bucket(bucket['value'])
			except errors.ServerSelectionTimeoutError:
				# log_err('Unable to connect to mongo!')
				sys.exit(1)

	def _parse_bucket(self, data):
		self._name = data['name']
		self._owner = data['ownerDisplayName']
		self._transient = data['transient']
		self._deleted = data['deleted']
		self._location = data['locationConstraint']
		self._created = data['creationDate']

	@property
	def name(self):
		self._pull_info()
		return self._name

	@property
	def owner(self):
		self._pull_info()
		return self._owner

	@property
	def created(self):
		self._pull_info()
		return self._created

	@property
	def transient(self):
		self._pull_info()
		return ('True', 'green') if self._transient else ('False', 'red')

	@property
	def deleted(self):
		self._pull_info()
		return self._deleted

	@property
	def location(self):
		self._pull_info()
		return self._location

def build_buckets(mongo, bkts):
	for b in bkts:
		yield ZenkoBucket(mongo, b)



import sys
from itertools import chain
from pprint import pprint

from pymongo import MongoClient, errors

from Crypto.Cipher import PKCS1_OAEP
from Crypto.Hash import SHA256
from Crypto.PublicKey import RSA
from collections import namedtuple
import base64
from .s3 import BackendWrapper
from .util import MONGO_BASE_HOST
from . import error
from .log import Log

_log = Log('orbit')

BACKEND_TYPES = {
	"location-mem-v1" : 'MEM',
	"location-file-v1": 'FILE',
	"location-azure-v1": 'AZURE',
	"location-aws-s3-v1": 'AWS',
	"location-gcp-v1": 'GCP',
	"location-b2-v1": 'BACKBLAZE B2',
	"location-wasabi-v1": 'WASABI',
	"location-do-spaces-v1": 'DIGITAL OCEAN',
	'location-scality-ring-s3-v1': 'SCALITY RING S3C',
	'location-scality-sproxyd-v1': 'SCALITY RING SPD',
	'location-ceph-radosgw-s3-v1': 'CEPH'
}

User = namedtuple('User', ['name', 'access_key', 'secret_key'])
Backend = namedtuple('Backend', ['name', 'type', 'access_key', 'secret_key', 'endpoint', 'bucket', 'transient'])

class ReplicationStream:
	_headings = ('name', 'source', 'prefix', 'destinations', 'enabled')
	_heading_color = 'cyan'
	def __init__(self, **kwargs):
		self._name = kwargs.get('name')
		self._stream_id = kwargs.get('streamId')
		self._enabled = kwargs.get('enabled', False)
		self._source = kwargs.get('source', {}).get('bucketName')
		self._prefix = kwargs.get('source', {}).get('prefix')
		self._destinations = [(l['name'], l['storageClass']) for l in kwargs.get('destination', {}).get('locations', [])]

	@property
	def destinations(self):
		return ','.join(n for n, c in self._destinations)

	@property
	def repr(self):
		color = 'green' if self._enabled else 'red'
		return self._name, self._source, self._prefix if self._prefix else '----', self.destinations, (str(self._enabled), color)

	@classmethod
	def headings(cls):
		return tuple((k.upper(), cls._heading_color) for k in cls._headings)

class OverlayConfig:
	_heading_color = 'cyan'
	_instance_headings = ('instance id',)
	_user_headings = ('username', 'account type', 'access key', 'canonicalId')
	_location_headings = ('name', 'type', 'access key', 'bucket', 'transient', 'endpoint')
	_endpoint_headings = ('hostname', 'location', 'builtin')
	_backend_check_headings = ('backend', 'bucket', 'transient', 'exists', 'owned')

	def __init__(self, helm_release = None, mongo = None, verbose = False):
		if mongo is None:
			if helm_release is None:
				raise Exception('You must provide either monge or helm_release!')
			self._host ='%s-%s'%(helm_release, MONGO_BASE_HOST)
		else:
			self._host = mongo
		self._client = self._build_client(self._host)
		self._verbose = verbose
		self._pulled = False
		self._pull_config()

	def _build_client(self, host):
		return MongoClient(host,
			connectTimeoutMS = 5000,
			serverSelectionTimeoutMS = 5000)

	@property
	def mongo(self):
		return self._host

	def _pull_priv_key(self):
		if not 'metadata' in self._client.database_names():
			raise error.DBNotFoundException('metadata')
		db = self._client['metadata']
		if not 'PENSIEVE' in db.collection_names():
					raise error.CollectionNotFoundException('PENSIEVE', 'metadata')
		col = db['PENSIEVE']
		auth = col.find_one({'_id': 'auth/zenko/remote-management-token'})
		if auth is None:
			raise NoOverlayPrivateKeyException
		return RSA.importKey(auth['value']['privateKey'])

	def _build_cypher(self):
		key = self._pull_priv_key()
		return PKCS1_OAEP.new(key, hashAlgo=SHA256.new(), label='')

	@property
	def cypher(self):
		return self._build_cypher()

	def decrypt(self, data):
		return self.cypher.decrypt(
			base64.b64decode(data)
		).decode('utf-8')

	def _pull_config(self):
		if not self._pulled:
			try:
				if not 'metadata' in self._client.database_names():
					raise error.DBNotFoundException('metadata')
				db = self._client['metadata']
				if not 'PENSIEVE' in db.collection_names():
					raise error.CollectionNotFoundException('PENSIEVE', 'metadata')
				col = db['PENSIEVE']
				version = col.find_one({'_id':'configuration/overlay-version'})
				if version is None:
					raise error.NoOverlayConfigException
				config = col.find_one({'_id': 'configuration/overlay/%s'%version['value']})
				if config is None:
					raise error.InvalidOverlayConfigVersionException('"%s"'%version['value'])
				self._pulled = True
				self._parse_config(config['value'])
				return True, None
			except errors.ServerSelectionTimeoutError:
				_log.error('Failed to connect to mongodb')
				raise error.MongoConnectionError(self.mongo)

	def _parse_config(self, config):
		self._instance_id = config['instanceId']
		self._users = config['users']
		self._locations = list(config['locations'].values())
		self._replication_streams = [ReplicationStream(**rs) for rs in config['replicationStreams']]
		self._endpoints = config['endpoints']

	@property
	def _repr_users(self):
		users = []
		for user in self._users:
			cid = user['canonicalId'] if self._verbose else '%s...%s'%(user['canonicalId'][:10], user['canonicalId'][-15:])
			ak = user['accessKey'] if self._verbose else '%s...%s'%(user['accessKey'][:6], user['accessKey'][-5:])
			users.append((user['userName'], user['accountType'], ak, cid))
		return users

	@property
	def _repr_locations(self):
		locations = []
		for loc in self._locations:
			details = loc.get('details', None)
			transient = ('True', 'green') if loc.get('isTransient', False) else ('False', 'red')
			if details:
				ak = details.get('accessKey', '----')
				ak = ak if self._verbose or ak == '----' else '%s...%s'%(ak[:6], ak[-5:])
				ep = details.get('endpoint', '----')
				bucket = details.get('bucketName', '----')
			else:
				ep, ak, bucket = '----', '----', '----'
			locations.append((loc['name'], BACKEND_TYPES.get(loc['locationType'], 'UNKNOWN'), ak, bucket, transient, ep))
		return locations

	@property
	def _repr_replication_streams(self):
		return [rs.repr for rs in self._replication_streams]

	@property
	def _repr_endpoints(self):
		eps = []
		for ep in self._endpoints:
			# builtin = ('True', 'magenta') if ep.get('isBuiltin', False) else 'False'
			eps.append((ep['hostname'], ep['locationName'], str('isBuiltin' in ep)))
		return eps

	@property
	def _repr_instace_id(self):
		return [((self._instance_id, 'green'),)]

	def headings(self, section):
		headings = getattr(self, '_%s_headings'%section, None)
		if headings is None:
			raise Exception('No headings found for %s!'%section)
		return tuple((k.upper(), self._heading_color) for k in headings)

	@property
	def repr(self):
		self._pull_config()
		rpr = {
			'instance id': self._repr_instace_id,
		}
		if self._users:
			rpr['users'] = [self.headings('user')] + self._repr_users
		if self._locations:
			rpr['locations'] = [self.headings('location')] + self._repr_locations
		if self._endpoints:
			rpr['endpoints'] = [self.headings('endpoint')] + self._repr_endpoints
		if self._replication_streams:
			rpr['replication streams'] = [ReplicationStream.headings()] + self._repr_replication_streams
		return rpr

	@property
	def users(self):
		for user in self._users:
			yield User(user['userName'], user['accessKey'], self.decrypt(user['secretKey']))

	@property
	def backends(self):
		self._pull_config()
		for backend in self._locations:
			transient = backend.get('isTransient', False)
			if not 'details' in backend or not backend.get('details'):
				yield Backend(
					backend['name'], BACKEND_TYPES.get(backend['locationType'], 'UNKNOWN'),
					'----', '----', '----', '----', transient
				)
				continue
			details = backend['details']
			if backend['locationType'] == 'location-file-v1' or backend['locationType'] == 'location-scality-sproxyd-v1':
				yield Backend(
					backend['name'], BACKEND_TYPES.get(backend['locationType'], 'UNKNOWN'),
					'----', '----', details.get('endpoint', 'None'), '----', transient
				)
			else:
				yield Backend(
					backend['name'], BACKEND_TYPES.get(backend['locationType'], 'UNKNOWN'),
					details['accessKey'], self.decrypt(details['secretKey']),
					details.get('endpoint', 'None'), details['bucketName'], transient
				)



from .util import WatchProcess, check_cmd, NO_PROBLEMS
import subprocess
import socket
import click
from .error import RequiredBinaryException
from .print import pad, print_info, TERM_WIDTH

class Service:
	_headings = ('HOSTNAME', 'CLUSTER IP', 'PORTS', 'AGE', 'STATUS')
	_heading_color = 'cyan'

	def __init__(self, line, check = False, verbose = False):
		self._ports = []
		self._check = check
		self._verbose = verbose
		self._parse_line(line)
		self._status = None

	def _parse_line(self, line):
		parts = [p for p in line.split() if p]
		self._host = parts[0]
		self._ip = parts[2]
		for p in parts[4].split(','):
			port, proto = p.split('/')
			self._ports.append((port, proto))
		self._age = parts[5]

	@property
	def status(self):
		if self._status is None:
			self._status = self.check()
		return self._status

	@property
	def endpoints(self):
		for p in self.ports:
			yield self._host, p

	@property
	def ports(self):
		for port, type in self._ports:
			yield port

	@property
	def pports(self):
		return ','.join('/'.join(pair) for pair in self._ports)

	@property
	def ip(self):
		return self._ip

	@property
	def age(self):
		return self._age

	@property
	def repr(self):
		return self._host, self._ip, self.pports, self._age, ('UP', 'green') if self.status else ('DOWN', 'red')

	def headings(self):
		return tuple((c, Service._heading_color) for c in Service._headings)

	def check(self):
		up = True
		for host, port in self.endpoints:
			if self._ip == 'None':
				up = False
			try:
				s = socket.create_connection((host, port))
			except Exception as e:
				if self._verbose:
					click.secho('Failed to connect to tcp://%s:%s with error %s'%(host, port, e), fg='red')
				up = False
		return up

class Pod:
	_headings = ('NAME', 'READY', 'RESTARTS', 'AGE', 'STATUS')
	_heading_color = 'cyan'
	def __init__(self, line):
		self._parse_line(line)

	def _parse_line(self, line):
		parts = [p for p in line.split() if p]
		self._name = parts[0]
		self._ready = parts[1]
		self._status = parts[2]
		self._restarts = parts[3]
		self._age = parts[4]

	@property
	def name(self):
		return self._name

	@property
	def ready(self):
		return tuple(int(i) for i in self._ready.split('/'))

	@property
	def status(self):
		return self._status

	@property
	def restarts(self):
		return int(self._restarts)

	@property
	def age(self):
		return self._age

	@property
	def repr(self):
		fg = 'green' if self._status == 'Running' or self.status == 'Completed' else 'red'
		restarts_color = 'green'
		if int(self._restarts) >= 5:
			restarts_color = 'yellow'
		if int(self._restarts) >= 10:
			restarts_color = 'red'
		return self.name, self._ready, (self._restarts, restarts_color), self._age, (self._status, fg)

	@classmethod
	def headings(cls):
		return tuple((c, cls._heading_color) for c in cls._headings)



class Status:
	def __init__(self, release, check_services = False, verbose = False, host = None):
		self._release = release
		self._services = []
		self._pods = []
		self._check_services = check_services
		self._host = host
		if self._host is None:
			self._host = ''
		else:
			self._host = '--host %s'%self._host
		self._verbose = verbose
		if not check_cmd('helm'):
			raise RequiredBinaryException('helm')

	def _helm_status(self):
		proc = WatchProcess('helm %s status %s'%(self._host, self._release), stdout = subprocess.PIPE, stderr = subprocess.STDOUT, env=dict(KUBECONFIG=''))
		proc.wait()
		raw = proc().stdout.readlines()
		lines = [x.decode('utf-8').strip() for x in raw]
		return lines

	def _parse_services(self, lines):
		for line in lines:
			self._services.append(Service(line, check = self._check_services, verbose = self._verbose))

	def _parse_pods(self, lines):
		for line in lines:
			self._pods.append(Pod(line))

	def _parse(self, lines):
		section_ids = {
			'v1/Service': self._parse_services,
			'v1/Pod': self._parse_pods
		}
		for i, line in enumerate(lines):
			seg = line.replace('==> ', '').replace('(related)', '')
			if seg in section_ids:
				for ei, eline in enumerate(lines[i + 1:]):
					if not eline:
						end = ei + i
						break
				section_ids[seg](lines[i + 2:end + 1])

	def pull_status(self):
		lines = self._helm_status()
		if len(lines) == 1 and 'Error' in lines[0]:
			return False
		self._parse(lines)
		return True

	@property
	def repr(self):
		if not self._verbose:
			with click.progressbar(self._services, label="Checking services") as service_list:
				services = [s.repr for s in service_list]
			with click.progressbar(self._pods, label = 'Checking pods    ') as pod_list:
				pods = [p.repr for p in pod_list]
		else:
			pods = [p.repr for p in self._pods]
			services = [s.repr for s in self._services]
		if not self._verbose:
			services = [s for s in services if not s[4] == ('UP', 'green')]
			if not len(services):
				services = [(('NO PROBLEMS', 'green'),)]
			# if len(services) == len(self._services):
			# 	click.echo('\n')
			# 	print_info(pad('Note: If executed outside the Zenko cluster, service checks ALWAYS return DOWN', TERM_WIDTH, align='center'), fg='yellow')
				# services = [(('Note: If executed outside the Zenko cluster service checks ALWAYS return DOWN', 'yellow'),)] + services
			pods = [p for p in pods if not p[4] == ('Running', 'green') and not p[4] == ('Completed', 'green')]
			if not len(pods):
				pods = [(('NO PROBLEMS', 'green'),)]
		services = [self._services[0].headings()] + services if len(services) > 1 else services
		pods = [Pod.headings()] + pods if len(pods) > 1 else pods
		return dict(services=services, pods=pods)



from .s3 import build_backends
import click
from .bucket import UserBuckets


def check_backends(overlay, verbose = False):
	yield overlay.headings('backend_check')
	for backend in build_backends(overlay.backends):
		transient = ('True', 'green') if backend.transient else 'False'
		if backend.builtin and verbose:
			yield backend.name, backend.bucket, transient, ('Builtin', 'magenta'), ('Builtin', 'magenta')
			continue
		if backend.exists is True:
			exists = ('True', 'green')
		elif backend.exists is False:
			exists = ('False', 'red')
		else:
			exists = ('----', 'yellow')
		# mine = ('True', 'green') if backend.mine is True else ('False', 'red')
		if backend.mine is True:
			mine = ('True', 'green')
		elif backend.mine is None:
			mine = ('----', 'yellow')
		else:
			mine = ('False', 'red')
		if not backend.exists or not backend.mine or verbose:
			yield backend.name, backend.bucket, transient, exists, mine

def check_buckets(overlay, buckets, verbose = False):
	backends = {b.name: b.exists for b in build_backends(overlay.backends)}
	transient_backends = [b.name for b in overlay.backends if b.transient]
	yield tuple([(x, 'cyan') for x in ['NAME', 'BACKEND', 'OWNER', 'TRANSIENT', 'VALID']])
	for bucket in buckets:
		if not bucket.deleted:
			valid = ('True', 'green') if bucket.location in backends and backends[bucket.location] else ('False', 'red')
			if not bucket.location in backends or not backends[bucket.location] or verbose:
				yield bucket.name, bucket.location, bucket.owner, str(bucket.location in transient_backends), valid

def check_bucket_users(overlay, buckets, verbose = False):
	pass






ORBIT_ERROR_MSG = 'This probably means that Cloudserver is having difficulty communicating with Orbit.'

class ZCheckBaseException(Exception):
	def __init__(self, *args, **kwargs):
		super().__init__(self._msg_tmpl.format(*args, **kwargs))


class RequiredBinaryException(ZCheckBaseException):
	_msg_tmpl = 'No binary name {} found in PATH\nYou must have {} installed to use this tool!\nPlease make sure {} is installed and available in you PATH'
	def __init__(self, cmd):
		super().__init__(*([cmd]*3))

class MongoConnectionError(ZCheckBaseException):
	_msg_tmpl = 'Unable to connect to mognodb at {}!'

class DBNotFoundException(ZCheckBaseException):
	_msg_tmpl = 'Unable to find {} database in mongodb!'

class CollectionNotFoundException(ZCheckBaseException):
	_msg_tmpl = 'Unable to find collection {} in db {} in mognodb!'

class NoOverlayConfigException(ZCheckBaseException):
	_msg_tmpl = 'No overlay configuration found!\n' + ORBIT_ERROR_MSG

class InvalidOverlayConfigVersionException(ZCheckBaseException):
	_msg_tmpl = 'Could not find overlay version {}\n' + ORBIT_ERROR_MSG

class NoOverlayPrivateKeyException(ZCheckBaseException):
	_msg_tmpl = 'Could not find Zenkos private key to access secrets!\n' + ORBIT_ERROR_MSG



from threading import Thread, Event
import os
import os.path
import subprocess
import shlex
import click

MONGO_BASE_HOST = 'mongodb-replicaset'
NO_PROBLEMS = [(('NO PROBLEMS', 'green'),)]

def which(name):
	for path in os.getenv("PATH").split(os.path.pathsep):
		full_path = path + os.sep + name
		if os.path.exists(full_path):
			return True
	return False

def complain_no_cmd(cmd, exit = True):
	click.secho('''No binary name %s found in PATH
You must have %s installed to use this tool!
Please make sure %s is installed and available in you PATH'''%tuple([cmd]*3),
	fg='red', bold=True)

def check_cmd(cmd):
	if not which(cmd):
		complain_no_cmd(cmd)
		return False
	return True

class WatchedProcess(Thread):
	"""
		A light wrapper around a Popen object
		all args are passed through to the Popen constructor
		2 additional keyword arguments are added
			on_exit
			on_error
		These should contain a callable object taking 1 arguement return_code
		on_exit will always be called when the process exits
		on_error will be called when the process exits with return_code != 0
	"""

	def __init__(self, *args, on_exit = None, on_error = None, **kwargs):
		super().__init__()
		self.daemon = True
		self._proc = None
		self._started = Event()
		self._args = args
		self._kwargs = kwargs
		self._on_exit = on_exit
		self._on_error = on_error
		self._proc = subprocess.Popen(*self._args, **self._kwargs)

	def __call__(self):
		"""for convenience return the popen object when called"""
		return self._proc

	def run(self):
		self._started.set()
		self._proc.wait()
		rc = self._proc.returncode
		if self._on_exit:
			self._on_exit(rc)
		if self._on_error and rc != 0:
			self._on_error(rc)

	def terminate(self):
		return self._proc.terminate()

	def kill(self):
		return self._proc.kill()

	@property
	def status(self):
		if self._proc:
			return self._proc.poll()

	def wait(self):
		self._started.wait()
		return self._proc.wait() if self._proc else None


def WatchProcess(cmd, start = True, **kwargs):
	wp = WatchedProcess(shlex.split(cmd), **kwargs)
	if start:
		wp.start()
	return wp

def mark_last(it):
	if not hasattr(it, '__next__'):
		it = iter(it)
	try:
		value = next(it)
	except StopIteration:
		pass
	else:
		for v in it:
			yield False, value
			value = v
		yield True, value



import logging
from logging import CRITICAL, ERROR, WARNING, INFO, DEBUG

levels = dict(
		critical = logging.CRITICAL,
		error = logging.ERROR,
		warning = logging.WARNING,
		info = logging.INFO,
		debug = logging.DEBUG
		)

LOGGERS = dict()

class Whitelist(logging.Filter):
	def __init__(self, *whitelist):
		self.whitelist = [logging.Filter(name) for name in whitelist]

	def filter(self, record):
		return any(f.filter(record) for f in self.whitelist)

class Blacklist(Whitelist):
	def filter(self, record):
		return not Whitelist.filter(self, record)

def formatter():
	return logging.Formatter(fmt='%(asctime)s %(name)s %(levelname)s: %(message)s',
		datefmt='%d-%m-%y %I:%M:%S %p')

def setupLogging(name = None, level = logging.DEBUG, whitelist = [], blacklist = [], logfile = False):
	# Create our root logger and set the log lvl
	rootLogger = logging.getLogger()
	rootLogger.setLevel(level)

	# Setup formatting
	# formatter = logging.Formatter(fmt=logfmt, datefmt=datefmt)

	# Add logging to stdout
	streamHandler = logging.StreamHandler()
	streamHandler.setFormatter(formatter())

	rootLogger.addHandler(streamHandler)

	# Setup log files and rotation if enabled
	if logfile:
			handler = logging.FileHandler(logfile)
			handler.setFormatter(formatter())
			rootLogger.addHandler(handler)

	if whitelist:
		for handler in logging.root.handlers:
			handler.addFilter(Whitelist(*whitelist))
	elif blacklist:
		for handler in logging.root.handlers:
			handler.addFilter(Blacklist(*blacklist))
	name = name if name else 'zcheck'
	baseLogger = rootLogger.getChild(name)
	LOGGERS[''] = baseLogger
	return baseLogger

# return a nested child of root
# levels are indicated in name by "."
# eg. "root.foo.bar"
def get_logger(root, name):
	if not '.' in name:
		return root.getChild(name)
	lvl = name.split('.')
	return get_logger(root.getChild(lvl.pop(0)), '.'.join(lvl))

def Log(name):
	baselogger = LOGGERS.get('', logging.getLogger())
	return get_logger(baselogger, name)

def log_on_error(logger, target_handler = None, flush_lvl = logging.ERROR, capacity = 250):
	if target_handler is None:
		target_handler = logging.StreamHandler()
		target_handler.setFormatter(formatter())
	handler = logging.handlers.MemoryHandler(capacity, flushLevel = flush_lvl, target = target_handler)

	def decorator(fn):
		def wrapper(*args, **kwargs):
			logger.addHandler(handler)
			try:
				return fn(*args, **kwargs)
			except Exception:
				raise
			finally:
				super(logging.handlers.MemoryHandler, handler).flush()
				logger.removeHandler(handler)
		return wrapper
	return decorator



from boto3 import Session
import botocore
from botocore.handlers import set_list_objects_encoding_type_url
from azure.storage.blob import BlockBlobService


class AzureClientError(Exception):
	def __init__(self, error):
		self._error = error

	@property
	def response(self):
		return dict(Error=dict(Code=self._error))

class ObjectStub:
	def __init__(self, k, v = None):
		def func(*args, **kwargs):
			pass
		setattr(self, k, v if v is not None else func)

class AzureResource:
	def __init__(self, service):
		self._service = service

	def Bucket(self, name):
		return AzureBucket(self._service, name)

	def resource(self, *args, **kwargs):
		return self

	@property
	def meta(self):
		os = ObjectStub('head_bucket', self._head_bucket)
		return ObjectStub('client', os)

	def _head_bucket(self, Bucket = None):
		if self._service.exists(Bucket):
			return True
		raise AzureClientError('404')

class AzureBucket:
	def __init__(self, service, name):
		self._service = service
		self._name = name

	@property
	def name(self):
		return self._name

	def create(self):
		return self._service.create_container(self._name)

	def delete(self):
		return self._service.delete_container(self._name)

	def put_object(self, Key = None, Body = None):
		if Key is None or Body is None:
			raise Exception
		return self._service.create_blob_from_bytes(
			self._name,
			Key,
			Body
		)

	def download_fileobj(self, key, f):
		blob = self._service.get_blob_to_bytes(
			self._name,
			key
		)
		if blob:
			f.write(blob.content)
			return True
		else:
			raise Exception('Failed to get blob %s/%s from azure!'%(self._name, key))

	@property
	def objects(self):
		def f():
			return self._service.list_blobs(self._name)
		return ObjectStub('all', f)

	def Versioning(self):
		return ObjectStub('suspend')

	def  delete_blob(self, name):
		self._service.delete_blob(self._name, name, delete_snapshots = True)



def _build_aws(backend):
	return Session(aws_access_key_id = backend.access_key,
					aws_secret_access_key = backend.secret_key)

def _build_gcp(backend):
	sesh = _build_aws(backend)
	sesh.events.unregister('before-parameter-build.s3.ListObjects', set_list_objects_encoding_type_url)
	return sesh

def _build_azure(backend):
	return AzureResource(
			BlockBlobService(
				account_name = backend.access_key,
				account_key = backend.secret_key
			)
		)

BUILDERS = {
	'GCP': _build_gcp,
	'AZURE': _build_azure
}

def build_client(backend):
	sesh = BUILDERS.get(backend.type, _build_aws)(backend)
	if backend.endpoint == 'None':
		endpoint = None
	elif backend.type == 'GCP':
		endpoint = 'https://storage.googleapis.com'
	else:
		endpoint = 'https://%s'%backend.endpoint
	return sesh.resource('s3', endpoint_url=endpoint)


class BackendWrapper:
	def __init__(self, backend):
		self._backend = backend
		self.builtin = self._backend.type == 'FILE' or self._backend.type == 'MEM'
		if not self.builtin:
			self._client = build_client(backend)
		self._mine = None
		self._exists = None
	@property
	def name(self):
		return self._backend.name

	@property
	def bucket(self):
		return self._backend.bucket

	@property
	def exists(self):
		if self.builtin:
			return True
		if self._backend.type == 'SCALITY RING SPD':
			self._mine = None
			return None
		if self._exists is not None:
			return self._exists
		try:
			self._client.meta.client.head_bucket(Bucket = self._backend.bucket)
			self._mine = True
			self._exists = True
			return True
		except (botocore.exceptions.ClientError, AzureClientError) as e:
			error_code = int(e.response['Error']['Code'])
			if error_code == 403:
				self._mine = False
				return True
			elif error_code == 404:
				return False

	@property
	def mine(self):
		if self.builtin:
			return True
		if self._backend.type == 'GCP' or self._backend.type == 'SCALITY RING SPD':
			return None
		if self._mine is None:
			if self.exists:
				return self._mine
			return False
		return self._mine

	@property
	def versioned(self):
		pass

	@property
	def transient(self):
		return self._backend.transient

def build_backends(backends):
	for backend in backends:
		yield BackendWrapper(backend)



import os
from .format import strip_color, color_text, find_col_widths
from functools import partial
import sys
from .util import mark_last
import re
from math import ceil
import click
ansi_escape = re.compile(r'\x1B\[[0-?]*[ -/]*[@-~]')


TERM_WIDTH, TERM_HEIGHT = os.get_terminal_size() if sys.stdin.isatty() else (120, 24)
TERM_WIDTH = min(TERM_WIDTH, 120)

control_chars = dict.fromkeys(range(32))

def pad(s, width, pad = ' ', align = 'left'):
	needed = pad * (width - len(ansi_escape.sub('', s)))
	if align == 'left':
		return s + needed
	elif align == 'right':
		return needed + s
	half = needed[:ceil(len(needed) / 2)]
	return half + s

def print_line(row, widths, ellipsize = False, align = 'left', **kwargs):
	colored = [color_text(c) for c in row]
	align = align if len(widths) > 1 else 'center'
	for last, col in mark_last(enumerate(colored)):
		i, text = col
		if not last or len(widths) == 1:
			click.echo(pad(text, widths[i], align = align) + '  ', nl = last, **kwargs)
		else:
			click.echo(text, nl = last, **kwargs)
	# click.echo(tmpl.format(*colored), **kwargs)

def print_header(s, pad = '=', fg = 'cyan', **kwargs):
	tmpl = ' %s ' if s else '%s'
	line = '{0:{fill}^{width}}'.format(tmpl%s.upper(), fill = pad, width=TERM_WIDTH)
	click.secho('\n' + line + '\n', fg = fg, **kwargs)

def print_section(section, **kwargs):
	numcols = len(section[0])
	widths = find_col_widths(section, numcols, TERM_WIDTH)
	# tmpl = ' '.join(['{:<%s}'%w for w in widths])
	align = 'center' if numcols == 1 else 'left'
	for row in section:
		print_line(row, widths, align = align)
	# map(render, section)

def print_sections(sections, **kwargs):
	for heading, section in sections.items():
		print_header(heading)
		print_section(section)

def print_error(msg):
	click.secho(msg, fg='red')

def print_info(msg, fg = 'cyan'):
	click.secho(msg, fg=fg)



from .util import WatchProcess
import subprocess
from functools import partial
import re

FORWARDED_PORTS = {}

PF_TMPL = 'kubectl port-forward {pod} {namespace} {local}:{remote}'
LIST_PODS = 'kubectl get pods --all-namespaces -o=jsonpath="{.items[*].metadata.name}"'

MONGO_REGEX = r'.*mongodb.*'
TILLER_REGEX = r'.*tiller.*'


def auto_forward(mongo = False, tiller = False, use_ctx = False):
	def outer(f):
		def inner(ctx, *args, **kwargs):
			close_mongo = False
			close_tiller = False
			conf = ctx
			if use_ctx: # To handle pass_context and pass_obj
				conf = ctx.obj
			if conf.auto_forward and mongo:
				mongo_pf = forward_mongodb() # Returns an existing PortForward if any
				close_mongo = mongo_pf.open() # Returns True if port has not been opened / False If its already open
			if conf.auto_forward and tiller:
				tiller_pf = forward_tiller()
				close_tiller = tiller_pf.open()
			ret = f(ctx, *args, **kwargs)
			if close_mongo:
				mongo_pf.close()
			if close_tiller:
				tiller_pf.close()
			return ret
		return inner
	return outer


class PortForward:
	def __init__(self, pattern, remote, local=None, namespace=None):
		self._pattern = pattern
		self._remote = remote
		self._local = local if local is not None else remote
		self._namespace = '--namespace %s'%namespace if namespace is not None else ''
		self._pod_name = None
		self._proc = None

	@staticmethod
	def list_pods():
		resp = WatchProcess(LIST_PODS, stdout=subprocess.PIPE)().communicate()
		return str(resp[0]).replace("'", '').split(' ')

	@property
	def pod(self):
		if self._pod_name is not None:
			return self._pod_name
		for pod in self.list_pods():
			if re.search(self._pattern, pod):
				self._pod_name = pod
				return pod

	def open(self):
		if self._proc is None:
			self._proc = WatchProcess(PF_TMPL.format(
				pod=self.pod,
				namespace=self._namespace,
				local=self._local,
				remote=self._remote),
				stdout=subprocess.DEVNULL,
				stderr=subprocess.DEVNULL)
			return True
		return False

	def close(self):
		if self._proc is not None:
			self._proc.terminate()
			return True
		return False


def forward_pod(pattern, port, local = None, namespace = None):
	if pattern in FORWARDED_PORTS:
		return FORWARDED_PORTS[pattern]
	pf = PortForward(pattern, port, local=local, namespace=namespace)
	FORWARDED_PORTS[pattern] = pf
	return pf


def get_port_forward(pattern = None):
	if pattern is None:
		return list(FORWARDED_PORTS.items())
	return FORWARDED_PORTS.get(pattern, None)


forward_mongodb = partial(forward_pod, MONGO_REGEX, 27017, 40420)
forward_tiller = partial(forward_pod, TILLER_REGEX, 44134, 40421, namespace = 'kube-system')



import click
from math import floor
from .util import mark_last
from io import StringIO
import termcolor

MAX_GAP = 10

def color_text(text):
	if isinstance(text, tuple):
		return termcolor.colored(*text)
	return text

def strip_color(line):
	if isinstance(line, tuple):
		return line[0]
	return line

def find_col_max_widths(section, numcols):
	max_per_col = []
	for col in range(numcols):
		longest = max([len(strip_color(l[col])) for l in section])
		max_per_col.append(longest)
	return max_per_col

def distribute_extra(widths, numcols, extra):
	per_col = floor(extra / numcols)
	widths = list(map(lambda x: x + per_col, widths))
	leftover = extra - (per_col * numcols)
	if leftover:
		min_len = min(widths)
		pos = widths.index(min_len)
		widths[pos] += leftover
	return widths

def find_col_widths(section,  numcols, width = 80):
	if numcols == 1:
		return [width]
	max_per_col = find_col_max_widths(section, numcols)
	naive_per_col = floor(width / numcols)
	widths = []
	for i in range(numcols):
		if max_per_col[i] > naive_per_col:
			widths.append(max_per_col[i])
		else:
			if naive_per_col - max_per_col[i] >= MAX_GAP:
				widths.append(max_per_col[i])
			else:
				widths.append(naive_per_col)
	return widths










def print_line(l, spaced = False, file = None, **kwargs):
	if spaced:
		click.secho('\n ' + l + '\n', **kwargs)
	else:
		click.secho(' '+ l, **kwargs)

def print_padded(line, width, pad = ' ', **kwargs):
	pad_width = int((width - len(line)) / 2)
	padding = pad * pad_width
	print_line('%s%s%s'%(padding, line, padding), **kwargs)

def print_header(s, width = 80, pad = '=', fg = 'cyan', level = 0, **kwargs):
	inner_width = width - (5 * level)
	inner = '{:{fill}^{width}}'.format(' %s '%s, width = inner_width, fill = pad)
	line = '{:^{width}}'.format(inner, width=width)
	print_line(line, fg = fg, spaced = True, **kwargs)

def print_colored_row(row, widths = None, total_width = None, **kwargs):
	if not widths:
		widths = [0] * len(row)
	if total_width:
		if len(row) == 1:
			extra_padding = int((total_width - sum(widths)) / 2)
		else:
			extra_padding = int((total_width - sum(widths)) / (len(row) - 1))
		if extra_padding < 0:
			extra_padding = 0
	else:
		extra_padding = 0
	for i, col in enumerate(row):
		if isinstance(col, tuple):
			line, color = col
		else:
			line = col
			color = None
		# print(i, len(row)-1, end='')
		if i == len(row) - 1:
			nl = True
		else:
			nl = False
		if i < len(row) - 1 or len(row) == 1:
			if len(row) == 1:
				print_line(left_pad(line, widths[i] + extra_padding), fg = color, nl = nl, **kwargs)
			elif widths:
				print_line(right_pad(line, widths[i] + extra_padding), fg = color, nl = nl, **kwargs)
			else:
				print_line(right_pad(line, extra_padding + 1), fg = color, nl = nl, **kwargs)
		else:
			print_line(line, fg = color, nl = nl, **kwargs)

def print_row(row, total_width = 80, **kwargs):
	cols = len(row)
	col_width = floor(total_width / cols)
	colored_row = [color_text(c)[:col_width] for c in row]
	tmpl = '{:<{width}}'
	# click.echo
	# click.echo(tmpl.format(*colored_row, width=col_width))
	for last, col in mark_last(colored_row):
		print_line(tmpl.format(col, width=col_width), nl=last, **kwargs)







def max_length(lines, col):
	m = 0
	for line in lines:
		if len(strip_color(line[col])) > m:
			m = len(strip_color(line[col]))
	return m

def left_pad(line, width, pad = ' '):
	return '%s%s'%(pad*(width - len(line)), line)

def right_pad(line, width, pad = ' '):
	return '%s%s'%(line, pad * (width - len(line)))

def print_columns(cols, **kwargs):
	for row in cols:
		print_row(row, **kwargs)

def print_sections(sections, width = 80, _level = 0, **kwargs):
	for heading, section in sections.items():
		print_header(heading.upper(), **kwargs, level = _level, width = width)
		if isinstance(section, dict):
			print_sections(section, width = width, _level = _level + 1, **kwargs)
		else:
			print_columns(section, total_width = width - 5, **kwargs)



import sys
from collections import namedtuple
from pprint import pprint

import click

from . import __version__ as zcheck_version
from . import error
from .bucket import UserBuckets
from .check import check_backends, check_buckets
from .forward import auto_forward
from .helm import Status as HelmStatus
from .log import Log
from .orbit import OverlayConfig
from .print import print_error, print_header, print_sections, pad, TERM_WIDTH
from .util import MONGO_BASE_HOST, NO_PROBLEMS

_log = Log('cli')

ZCConf = namedtuple('ZCConf', ('mongo', 'helm_release', 'output', 'verbose', 'auto_forward', 'helm_host'))
NO_HELM_REL = ['help', 'version']
REQ_HELM_REL = ['k8s']



# Root
@click.group()
@click.option('--mongo', default = None, type = str, help = 'Override the default mongo host:port')
@click.option('--helm-release', '-r', type = str, help ='The release name helm was installed under')
@click.option('--helm-host', type = str, default = None, help = 'Override default tiller host')
@click.option('--output', '-o', type = click.File(mode='w'))
@click.option('--auto-forward', is_flag = True, help='With a correctly configured kubectl, auto-forward ports need for cluster introspection. (Overrides --mongo & --helm-host')
@click.option('--verbose', '-v', is_flag=True, help="Enable verbose output, WARNING: This will print your access keys to the terminal or file!")
@click.pass_context
def zcheck(ctx, **kwargs):
	subcmd = ctx.invoked_subcommand
	if kwargs.get('helm_release') is None:
		if subcmd in REQ_HELM_REL or (kwargs.get('mongo') is None and subcmd not in NO_HELM_REL):
			raise click.UsageError('Missing option "--helm-release" / "-r"', ctx=ctx)
	if 'output' not in kwargs:
		kwargs['output'] = None
	if kwargs.get('helm_host') and not ':' in kwargs.get('helm_host'):
		kwargs['helm_host'] = '%s:44134'%kwargs.get('helm_host')
	if kwargs.get('mongo') and not ':' in kwargs.get('mongo'):
		kwargs['mongo'] = '%s:27017'%kwargs.get('mongo')
	if kwargs.get('auto_forward'):
		kwargs['mongo'] = 'localhost:40420'
		kwargs['helm_host'] = 'localhost:40421'
	ctx.obj = ZCConf(**kwargs)

#help
@zcheck.command(help = 'Access usage help for commands')
@click.argument('command', required = False)
@click.pass_context
def help(ctx, command):
	if command:
		cmd = zcheck.get_command(ctx, command)
		if cmd:
			click.echo(cmd.get_help(ctx))
		else:
			click.secho('%s is not a valid command!'%command, fg='red')
	else:
		click.echo(zcheck.get_help(ctx))

# version
@zcheck.command(help='Print version')
def version():
	click.echo('zcheck', nl = False)
	click.secho('\tv%s'%zcheck_version, fg='green')

# k8s
@zcheck.command('k8s', help = 'Checks related to the kubernetes cluster')
@click.pass_obj
@auto_forward(tiller = True)
# @click.option('--check-services', '-c', is_flag = True, help = 'Attempt to connect to defined services and report their status')
def k8s(conf):
	try:
		hs = HelmStatus(conf.helm_release, verbose = conf.verbose, host  = conf.helm_host)
	except error.RequiredBinaryException as e:
		_log.error(str(e))
		_log.exception(e)
		print_error(str(e))
		sys.exit(1)

	if hs.pull_status():
		click.echo('\n')
		print_info(pad('Note: If executed outside the Zenko cluster, service checks ALWAYS return DOWN', TERM_WIDTH, align='center'), fg='yellow')
		print_sections(hs.repr, width=100, file=conf.output)
	else:
		print_error('Unable to retrieve helm status!')

# orbit
@zcheck.command('orbit', help = "Checks related to Orbit's configuration")
@click.pass_obj
@auto_forward(mongo = True)
def orbit(conf):
	try:
		oc = OverlayConfig(helm_release = conf.helm_release, mongo = conf.mongo, verbose=conf.verbose)
	except error.ZCheckBaseException as e:
		_log.error(str(e))
		_log.exception(e)
		print_error(str(e))
		sys.exit(1)
	print_sections(oc.repr, file=conf.output)

# backend
@zcheck.command('backends', help = 'Check backend buckets for existence and configuration')
@click.option('--deep', '-d', is_flag = True, help = 'Enable deep checking. Check every zenko bucket for its respective backend bucket')
@click.pass_context
@auto_forward(mongo = True, use_ctx = True)
def backends(ctx, deep):
	conf = ctx.obj
	try:
		oc = OverlayConfig(helm_release = conf.helm_release, mongo = conf.mongo, verbose=conf.verbose)
	except error.ZCheckBaseException as e:
		_log.error(str(e))
		_log.exception(e)
		print_error(str(e))
		sys.exit(1)
	checked = list(check_backends(oc, conf.verbose))
	if len(checked) == 1:
		checked = NO_PROBLEMS
	print_sections({'backends': checked})
	if deep:
		ctx.invoke(buckets)

# Check zenko buckets
@zcheck.command('buckets', help = 'Check every Zenko bucket for its respective backing bucket')
@click.pass_obj
@auto_forward(mongo = True)
def buckets(conf):
	try:
		oc = OverlayConfig(helm_release = conf.helm_release, mongo = conf.mongo, verbose=conf.verbose)
	except error.ZCheckBaseException as e:
		_log.error(str(e))
		_log.exception(e)
		print_error(str(e))
		sys.exit(1)
	ub = UserBuckets(helm_release=conf.helm_release, mongo = conf.mongo)
	checked = list(check_buckets(oc, ub.buckets, conf.verbose))
	if len(checked) == 1:
		checked = NO_PROBLEMS
	print_sections({'buckets': checked})

# Do EVERYTHING!!
@zcheck.command('checkup', help = 'Run all checks and tests')
@click.pass_context
@auto_forward(mongo = True, use_ctx = True)
def checkup(ctx):
	if not ctx.obj.helm_release:
		raise click.UsageError('Missing option "--helm-release" / "-r"', ctx=ctx)
	ctx.invoke(orbit)
	print_header('')
	ctx.invoke(k8s)
	ctx.invoke(backends)
	ctx.invoke(buckets)



__version__ = '0.1.4'


from .cli import zcheck
def entry():
	from . import log

	log.setupLogging('zcheck',
		level = log.CRITICAL,
		blacklist = [
			'azure.storage.common.storageclient'
		])
	zcheck()
